library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity control is
    port ( 
		reset: in std_logic; -- resets PC;
		clk: in std_logic; 
		-- input flags from datapath;
    V: in std_logic; --overflow flag
    C: in std_logic; --carry flag
    N: in std_logic; --negative flag
    Z: in std_logic; --zero flag 
		busA: in std_logic_vector(15 downto 0); --bus A
		-- output signals that represent fields of the control word
		-- which is used to control the datapath;
		DA: out std_logic_vector(2 downto 0); --address of destination register in register file 
		BA: out std_logic_vector(2 downto 0); --address of source 2 register
		AA: out std_logic_vector(2 downto 0); --address of source 1 register	
		MB: out std_logic; -- mux B	
		FS: out std_logic_vector(3 downto 0); -- FS part of the opcode
		MD: out std_logic; -- mux D	
		RW: out std_logic; -- write of register file 
		MW: out std_logic; -- write of data memory 
		-- "constant in" signal used by datapath;
		zero_fill: out std_logic_vector(15 downto 0)
    );
end control;


architecture control_arch of control is

component instruction_memory is port ( 
  addr_in : in std_logic_vector(15 downto 0);
  instr_out : out std_logic_vector(15 downto 0) );
end component;

-- signals used by the branch control 
-- if there is to be a jump or branch, PL is 1 loading the PC;
-- for PL = 0, the PC should simply be incremented;
signal PL: std_logic;
-- with PL = 1, JB = 1 calls for a jump and JB = 0 calls for 
-- a conditional branch;
signal JB, BC: std_logic;
signal JUMP: std_logic;	  
signal BRANCH: std_logic; --status bit that is the condition for the branch is selected;	 
-- PC related signals
signal PC, PC_next: std_logic_vector(15 downto 0); --program counter register; 
signal PC_offset: std_logic_vector(15 downto 0);
-- Note: next one helps to stop the execution;
signal STOP_EXECUTION: std_logic; 
-- output of the instruction memory
signal INSTRUCTION : std_logic_vector(15 downto 0); 


begin 
	
	
-- comb. logic that implements the logic for jump and 
-- branch signals; see Mano's book page 472;
-- Note :C, V not used in this control unit;
process (N, Z, PL, JB, BC) is 
begin 
	if (PL = '1') then -- we know we have a jump or branch
		if (JB = '1') then 
			JUMP <= '1'; -- a jump occurs if bit 13 of instruction is 1; 
			BRANCH <= '0';
		else -- a conditional branch occurs in this case;		
			JUMP <= '0';
			if (BC = '1') then
				BRANCH <= N;
			else 
				BRANCH <= Z;
			end if;
		end if;	
	else
		JUMP <= '0';
		BRANCH <= '0';
	end if;
end process;

-- comb. logic to set the PC offset for a branch (see Mano's book page 469); 
-- sign extended thing
--		  | 
--	      | PC_offset	| 1
--		  --------------- 
--	   ______	|
--	  |	  PC |	|
--    |       +  	
--    |       |            | busA <-- value stored in register R[SA]	
--    |       --------------
--    |             | PC_next     <-- input to PC register
--	  |		    REGISTER	      <-- clk
--	  |_____________| PC		  <-- out of PC register
--                  |             <-- to inout of instruction memory
offset_logic: process (INSTRUCTION) is
begin 	
	-- form offset from split address fields of the instruction;
	PC_offset(5 downto 3) <= INSTRUCTION(8 downto 6); --AD left;
	PC_offset(2 downto 0) <= INSTRUCTION(2 downto 0); --AD right;
	-- then sign extend it;
	for i in 15 downto 6 loop
    PC_offset(i) <= INSTRUCTION(8); -- basically PC_offset(5)
  end loop;
end process; 


-- PC register;	the only one that uses the clk signal;
PC_register: process (reset, clk) is 
begin
	if (reset = '1') then  
  	PC <= (others => '0'); -- reset PC
	elsif rising_edge(clk) then	 
		PC <= PC_next;
	end if;
end process;   
-- comb. logic to set the actual input ("next" value) of PC register
-- either use busA or add the offset or simply increment PC (i.e., add 1);
PC_logic: process (JUMP, BRANCH, PC) is
begin 
	if (STOP_EXECUTION = '1') then
		PC_next <= PC; -- freeze it;
	else -- actual logic here;
		if (JUMP = '1') then
			PC_next <= busA;
		else
			if (BRANCH = '1') then   
				PC_next <= std_logic_vector(signed(PC) + signed(PC_offset));
			else  
				PC_next <= std_logic_vector(unsigned(PC) + 1);
			end if;
		end if;	
	end if;
end process; 
-- always check if encoutered end of execution mark;
STOP_EXECUTION <= '1' when INSTRUCTION = X"FFFF" else '0';
   
		
-- instantiate the instruction memory; read only memory; 
-- we use only a memory with only 16 entries, only 4 bits are enough
-- to address each mem location; remaining bits are not used
-- in this implementation;
instruction_memory_unit: instruction_memory port map(
  addr_in => PC,
  instr_out => INSTRUCTION); 


-- Zero fill unit; 
zero_fill <= "0000000000000" & INSTRUCTION(2 downto 0);
 
	
-- comb. logic to implement instruction decoder; Mano's book page 473;
-- next signals are actual outputs of the control unit entity;
-- they control the datapath;
DA <= INSTRUCTION(8 downto 6);
AA <= INSTRUCTION(5 downto 3);	
BA <= INSTRUCTION(2 downto 0); 
MB <= INSTRUCTION(15);
FS(3 downto 1) <= INSTRUCTION(12 downto 10); 
FS(0) <= INSTRUCTION(9) AND (NOT PL); 
MD <= INSTRUCTION(13);
RW <= NOT INSTRUCTION(14);
MW <= INSTRUCTION(14) AND (NOT INSTRUCTION(15));
-- next signals are used here internally only, by the branch control;
PL <= INSTRUCTION(14) AND INSTRUCTION(15); 
JB <= INSTRUCTION(13);
BC <= INSTRUCTION(9); 
	
end control_arch;
