-- CopyLeft :-) Cristinel Ababei (cababei@buffalo.edu), December 2012
-- this is my version of VHDL implementation of the single-cycle computer (SCC)
-- described in Chapter 9 of:
-- [] M. Morris Mano and Charles Kime, Logic and Computer Design Fundamentals, 
-- Pearson Prentice Hall, 4th Edition, 2008.
-- I tested it in software and hardware and it seems to work fine for a simple program;
-- if you find errors or have improvements please send me an e-mail or a check 
-- in the amount of $1 :-).	
	
library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity single_cycle_cpu is -- this is the top-level entity of the SCC;
	port(
    reset: in STD_LOGIC;
    clk_100MHz : in std_logic; -- 100MHz clk, FPGA's external oscillator
    leds: out STD_LOGIC_VECTOR(7 downto 0)
  );
end single_cycle_cpu;

architecture structural of single_cycle_cpu is 

component ck_divider is
  Port ( 
    CK_IN: in STD_LOGIC;
    CK_OUT: out STD_LOGIC
  );
end component;

component control is
    port ( 
		reset: in std_logic; -- resets PC;
		clk: in std_logic; 
		-- input flags from datapath;
    V: in std_logic; --overflow flag
    C: in std_logic; --carry flag
    N: in std_logic; --negative flag
    Z: in std_logic; --zero flag 
		busA: in std_logic_vector(15 downto 0); --bus A
		-- output signals that represent fields of the control word
		-- which is used to control the datapath;
		DA: out std_logic_vector(2 downto 0); --address of destination register in register file 
		BA: out std_logic_vector(2 downto 0); --address of source 2 register
		AA: out std_logic_vector(2 downto 0); --address of source 1 register	
		MB: out std_logic; -- mux B	
		FS: out std_logic_vector(3 downto 0); -- FS part of the opcode
		MD: out std_logic; -- mux D	
		RW: out std_logic; -- write of register file 
		MW: out std_logic; -- write of data memory 
		-- "constant in" signal used by datapath;
		zero_fill: out std_logic_vector(15 downto 0) );
end component;

component datapath is
    port ( 
		reset: in std_logic; -- resets PC;
		clk: in std_logic; 	 
		-- outputs;
    V: out std_logic; --overflow flag
    C: out std_logic; --carry flag
    N: out std_logic; --negative flag
    Z: out std_logic; --zero flag 
		busA: buffer std_logic_vector(15 downto 0); --bus A	 
		busB: buffer std_logic_vector(15 downto 0); --bus B	 		
		-- inputs;	 
		dataIn: in std_logic_vector(15 downto 0); --bus A	 
		DA: in std_logic_vector(2 downto 0); --address of destination register in register file 
		BA: in std_logic_vector(2 downto 0); --address of source 2 register
		AA: in std_logic_vector(2 downto 0); --address of source 1 register	
		MB: in std_logic; -- mux B	
		FS: in std_logic_vector(3 downto 0); -- FS part of the opcode
		MD: in std_logic; -- mux D	
		RW: in std_logic; -- write of register file 
		MW: in std_logic; -- write of data memory 
		-- "constant in" signal used by datapath;
		zero_fill: in std_logic_vector(15 downto 0) );
end component;

component data_memory is
    port ( 
		reset : in std_logic;
    clk : in std_logic;
    MW : in std_logic;
    data_in : in std_logic_vector(15 downto 0);
    addr_in : in std_logic_vector(15 downto 0);
    data_out : out std_logic_vector(15 downto 0) );
end component;

-- connecting signals; 
signal addressOut : std_logic_vector(15 downto 0); --goes from datapath to data mem
signal dataOut : std_logic_vector(15 downto 0); --goes from datapath to data mem
signal dataIn : std_logic_vector(15 downto 0); --comes from data mem to datapath
signal constantIn : std_logic_vector(15 downto 0); --comes from control to datapath
signal busA : std_logic_vector(15 downto 0); --common to datapath and control
signal DA: std_logic_vector(2 downto 0); --address of destination register in register file 
signal BA: std_logic_vector(2 downto 0); --address of source 2 register
signal AA: std_logic_vector(2 downto 0); --address of source 1 register	
signal MB: std_logic; --mux B	
signal FS: std_logic_vector(3 downto 0); -- FS part of the opcode
signal MD: std_logic; --mux D	
signal RW: std_logic; --write of register file 
signal MW: std_logic; --write of data memory
signal V: std_logic; --overflow flag
signal C: std_logic; --carry flag
signal N: std_logic; --negative flag
signal Z: std_logic; --zero flag 		
signal clk: std_logic;
	
begin

CLOCK: ck_divider port map (clk_100MHz, clk);
-- at the end of the execution of current program this
-- should display 7;
-- Note: you should change how leds are driven based on what 
-- your program does; to make it more elegant leds should be driven
-- through a regsiter/buffer of its own...
leds <= dataOut(7 downto 0); 

CONTROL_UNIT: control port map (
	reset => reset,
	clk => clk,
  V => V,
  C => C,
  N => N,
  Z => Z,
	busA => addressOut, --driven by busA of datapath
	DA => DA,
	BA => BA,
	AA => AA,
	MB => MB,
	FS => FS,
	MD => MD,
	RW => RW,
	MW => MW,
	zero_fill => constantIn	);						   
	
DATAPATH_UNIT: datapath port map (
	reset => reset,
	clk => clk,	 
  V => V,
  C => C,
  N => N,
  Z => Z,
	busA => addressOut, --addressOut bus comes from busA and goes to data memory as address and to control as busA 
	busB => dataOut, --dataOut bus comes from busB and goes to data input of data memory 	
	dataIn => dataIn, --dataIn bus connects output of data memory here to datapath 
	DA => DA,
	BA => BA,
	AA => AA,
	MB => MB,
	FS => FS,
	MD => MD,
	RW => RW,
	MW => MW,
	zero_fill => constantIn	);

DATA_MEMORY_UNIT: data_memory port map ( 
	reset => reset,
  clk => clk,
  MW => MW,
  data_in => dataOut, -- out from datapath 
  addr_in => addressOut, -- out from datapath 
  data_out => dataIn ); -- input to datapath 

end structural;
